import bempp.api
from bempp.api.operators.boundary import helmholtz, sparse
from bempp.api.operators.potential import helmholtz as helmholtz_potential
from bempp.api.linalg import gmres
import numpy as np
from matplotlib import pyplot as plt
from scipy.interpolate import LinearNDInterpolator

from math import fabs
from math import pi
from math import sqrt
import meshio
import numpy
import sys

rho_0 = 1.21      # Density of air in kg/m^3
c_0 = 343         # Speed of sound in m/s 
f = 500           # Driving frequency in Hz
omega = 2*pi*f    # Angular frequency in rads/s
k = omega / c_0   # Wavenumber in rads/m
U_cone = 0.001    # Amplitude of the time-harmonic velocity of the cone in m/s
p_ref = 2e-5*sqrt(2)  # dB SPL ref (Pa)

g = meshio.read('panel.msh')
points = g.points
bemPoints = numpy.array([[p[0] for p in points], [p[1] for p in points], [p[2] for p in points]])
els = g.cells_dict['triangle']
bemEls = numpy.array( [[p[0] for p in els], [p[1] for p in els], [p[2] for p in els]] )
reData = numpy.array(g.point_data['deflection 1'])
imData = g.point_data['deflection im 1']
coeff = numpy.array([reData[i] + 1j * imData[i] for i in range(len(reData))])
tol = 1.0e-3

elData = []
for e in els:

    pi1 = e[0]
    pi2 = e[1]
    pi3 = e[2]

    pd1 = reData[pi1] + 1j * imData[pi1]
    pd2 = reData[pi2] + 1j * imData[pi2]
    pd3 = reData[pi3] + 1j * imData[pi3]

    elData.append((pd1+pd2+pd3)/3)

coefficients = numpy.array(elData)

grid = bempp.api.Grid(bemPoints, bemEls)

spaceP = bempp.api.function_space(grid, "P", 1)
spaceU = bempp.api.function_space(grid, "DP", 0)

u_total = bempp.api.GridFunction(spaceU, coefficients = coefficients) #fun=u_total_callable)
u_total.plot(transformation='imag') # transformation can be ‘real’, ‘imag’, ‘abs’, ‘log_abs’ or ‘abs_squared’ or a callable object

#sys.exit(0)


# Boundary operators
identity = sparse.identity(spaceP, spaceP, spaceP)
double_layer = helmholtz.double_layer(spaceP, spaceP, spaceP, k)
single_layer = helmholtz.single_layer(spaceU, spaceP, spaceP, k)


p_total, info = gmres(double_layer - 0.5 * identity,
                      1j * omega * rho_0 * single_layer * u_total, tol=1E-5)

#plot the total pressure on speaker boundary
p_total.plot(transformation='real') 


def plot_response_along_z():
    z = np.linspace(-4.0, 4.0, 41)
    points = np.array([np.zeros(len(z)), np.zeros(len(z)), z])
    pZ = helmholtz_potential.double_layer(spaceP, points, k) * p_total  \
        - 1j * omega * rho_0 * helmholtz_potential.single_layer(spaceU, points, k) * u_total
        
    plt.plot(z, np.real(pZ.ravel()), label='Real')
    plt.plot(z, np.imag(pZ.ravel()), label='Imag')
    plt.plot(z, np.abs(pZ.ravel()), label='Abs')
    plt.legend()
    plt.xlabel('z (m)')
    plt.ylabel('p_total (Pa)')
    plt.show()

def plot_directivity():
    theta = np.linspace(0, 2 * np.pi, 361)
    R_far_field = 4.0       # Evalution radius

    ## Compute pattern in XZ plane:
    points = R_far_field * np.array([np.sin(theta), np.zeros(len(theta)), np.cos(theta)])
    p_XZ = helmholtz_potential.double_layer(spaceP, points, k) * p_total  \
        - 1j * omega * rho_0 * helmholtz_potential.single_layer(spaceU, points, k) * u_total
    SPL_XZ = 20*np.log10(abs(p_XZ.ravel())/p_ref)
        
    ## Compute pattern in YZ plane:
    points = R_far_field * np.array([np.zeros(len(theta)), np.sin(theta), np.cos(theta)])
    p_YZ = helmholtz_potential.double_layer(spaceP, points, k) * p_total \
        - 1j * omega * rho_0 * helmholtz_potential.single_layer(spaceU, points, k) * u_total
    SPL_YZ = 20*np.log10(abs(p_YZ.ravel())/p_ref)
    
    plt.polar(theta, SPL_XZ, label='Vertical (XZ)')
    plt.polar(theta, SPL_YZ, label='Horizontal (YZ)')
    plt.legend()
    #plt.ylim(85, 95)
    plt.show()


plot_response_along_z()
plot_directivity()


def plot_exterior():
    # Make planes to plot pressure
    plotting_grid = bempp.api.import_grid('PlottingFaces.msh')
    #plotting_grid.plot()

    # 
    plotting_space = bempp.api.function_space(plotting_grid, "P", 1, include_boundary_dofs=True)
    double_layer_plotting = helmholtz.double_layer(spaceP, plotting_space, plotting_space, k)
    single_layer_plotting = helmholtz.single_layer(spaceU, plotting_space, plotting_space, k)
    p_plotting = double_layer_plotting * p_total - 1j * omega * rho_0 * single_layer_plotting * u_total
    p_plotting.plot(transformation='real')


    # Merge the two meshes
    grid_all = bempp.api.grid.union([grid, plotting_grid])

    # Create a space including all nodes
    space_all = bempp.api.function_space(grid_all, "P", 1, include_boundary_dofs=True)

    # Concatenate the coefficients
    all_coeffs = np.concatenate((p_total.coefficients, p_plotting.coefficients))

    # Convert to Sound Pressure Level
    nodes_SPL = 20 * np.log10(abs(all_coeffs) / p_ref)

    # Pack into a Bempp Grid Function
    plot_me = bempp.api.GridFunction(space_all, coefficients=nodes_SPL)

    # Plot
    plot_me.plot()